!+bug8
! Module: rgssources_mod
!
! Authors:	Peter Videler - ESTEC/SOC
!		Ian Stewart - SSC/LUX
!          
! Description: Construct table with source positions to be used in RGS pipeline.
! 
! Comments:
!
!	TODO:
!	- Check out if some things can be replaced by cal calls.
!	- test old list for empty labels?
!	- even more rigorous testing of old list??
!	- quit under any circumstances if old rgs list doesn't pass checks.
!	- make instrument param mandatory.
!	- write EXPOSURE kwd - see JO mail of 25 April.
!
! $Id: rgssources_mod.f90,v 1.5 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use param  
  use errorHandling
  use oal
  use rgslib
  use rgscelestoffset_mod
!  use datafilter

  implicit none 

! Globals:
  real(kind=double) ::&
!    pi,&
    twoPi,&
    piOn180,&
    rootPi,&
    root2,&
    root3

  integer(int16), parameter ::&
    labelsize = 20,&
    myInt16null = -999,&
    epicBandNum = 2
  character(len=labelsize) ::& ! These are defined in the sub myinit.
    proposalLabel,&
    onAxisLabel
  integer(int32), parameter ::&
    myInt32null = -999
  real(kind=single), parameter ::&
    myReal32null = -999.0
  type source_t
    integer(kind=int16) ::&
      index
    integer(int32) ::&
      epicFileAttrNum ! has to be 32 rather than 16 so can set an integer null using the dal call setNullValue.
    character(len=labelsize) ::&
      label
    real(kind=single) ::&
      raDeg,&
      decDeg,&
      rate,&
      deltaDispArcmin,&
      deltaXdspArcmin,&
      fovRdeg,&
      fovPhiDeg,&
      confusion
    logical(kind=bool) ::&
      process,&
      bkgExclude,&
      fixedOnSky
  endType source_t

! Command-line parameters:
  logical ::&
    withEpicSet,&
    addUserSrc,&
    changePrimeSrc,&
    changeAttitude,&
    clobberOnLabelClash,&
    setUserSrcToPrime
  character(len=255) ::&
    oldListSetName,&
    newListSetName,&
    epicSrcSetName,&
    atthkSetName,&
    primeSrcExpression
  character(len=9) ::&
    userSrcStyle,&
    primeSrcStyle,&
    attitudeStyle
  character(len=20) ::&
    instrument,&
    fileMode
  character(len=labelsize) ::&
    userSrcLabel,&
    primeSrcLabel
  character(len=labelsize-5) ::&
    epicLabelPrefix
  real(kind=single) ::&
    userSrcRate,&
    userSrcRaDeg,&
    userSrcDecDeg,&
    userSrcXdspArcmin,&
    userSrcDispArcmin,&
    userAttitudeRaDeg,&
    userAttitudeDecDeg,&
    userAttitudeAposDeg
  integer(kind=int16) ::&
    primeSrcIndex

  private 

  public rgssources

contains 

  !---------------------------------------------------------------------
  subroutine rgssources
  !---------------------------------------------------------------------

    implicit none 

    integer(int16) :: numOldListSrcs, numNewEpicSrcs, maxNumSrcs, actualNumSrcs, newPrimeSrcIndex
    type(source_t), dimension(:), allocatable :: newListSrcData!, oldListSrcData, newEpicSrcData
    type(DataSetT) :: oldListSet, epicSrcSet, newListSet
    type(TableT) :: oldListTab, epicSrcTab, newListTab
    type(SpacecraftAttitudeType) :: attitude          ! typedef from oal
    character(len=8) ::&
      epicExpIdAttrName,&
      epicExpId,&
      epicContentAttrName,&
      epicBandAttrName,&
      epicFilterAttrName,&
      epicFilter
    character(len=80) ::&
      epicContent
    integer(int16) :: newEpicAttrNum
    integer(int32) :: instrumentId
    character(len=5) :: instrumentName
!integer :: src

    !-------------------------------------------------------------------
    ! Start:

    call myinit
    call readParameters

    if (fileMode .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      instrumentName = stringAttribute(oldListSet, 'INSTRUME')
      call release(oldListSet)
      instrumentId = instId(instrumentName)

    else
!      instrument = ucase(instrument)
call ucase(instrument)
      if (instrument .eq. 'R1' .or. instrument .eq. 'RGS1') then
!        instrument = 'rgs1'
instrumentId = RGS1
      else&
      if (instrument .eq. 'R2' .or. instrument .eq. 'RGS2') then
!        instrument = 'rgs2'
instrumentId = RGS2
      else if (instrument .eq. 'NULLDEFAULT') then
call error('defaultInstrument', '')
! NOTE!!! This error dump relies for some extent on the parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
      else
call error('badInstrument', '')
! Note: make no restrictions on this parameter in .par file. There are too many possibilities.
      end if

      instrumentName = instId(instrumentId)
    end if

    call getPointing(attitude, instrumentId)

    numOldListSrcs = 0
    if (fileMode .eq. 'modify') then
      ! Open oldListSet and get number of rows:
      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      call release(oldListSet)
    end if

    numNewEpicSrcs = 0
    if (withEpicSet) then
      ! Open epicSrcSet and get number of rows:
      epicSrcSet = dataSet(epicSrcSetName, READ)

      epicExpId   = stringAttribute(epicSrcSet, 'INSTRUME')
      epicContent = stringAttribute(epicSrcSet, 'CONTENT')
      epicFilter  = stringAttribute(epicSrcSet, 'FILTER')

      epicSrcTab = table(epicSrcSet, 'SRCLIST', READ)
      numNewEpicSrcs = numberOfRows(epicSrcTab)
      call release(epicSrcSet)
    end if

    maxNumSrcs = numOldListSrcs + numNewEpicSrcs + 2 ! 2 for the proposal and
    ! onaxis sources.

    if (addUserSrc) maxNumSrcs = maxNumSrcs + 1

if (maxNumSrcs .le. 0) call error('bug0', '')

!print *, attitude%ra, attitude%dec, maxNumSrcs, instrumentName
!call error('','')
    allocate(newListSrcData(maxNumSrcs))
    call fillOutputSrcList(newListSrcData, maxNumSrcs, attitude&
    , actualNumSrcs, newEpicAttrNum)

!print *, 'after fillOutputSrcList'
!do src = 1, actualNumSrcs
!print *, src, newListSrcData(src)%label, len(newListSrcData(src)%label)&
!, len(trim(newListSrcData(src)%label))
!end do
!***** perform collision filtering or messaging

    newPrimeSrcIndex = calculatePrimeSourceIndex(newListSrcData, actualNumSrcs)
if (newPrimeSrcIndex .le. 0) call error('bug1', '') ! to pick up bugs in calculatePrimeSourceIndex(). Eventually this test (and the default setting line in the sub) could probably be dispensed with.

!print *, newPrimeSrcIndex

    call calculateConfusion(newListSrcData, actualNumSrcs, newPrimeSrcIndex)

!***** check now that no columns have nulls in them except EPIC_FILE.

!call dumpSrcData(newListSrcData, int(2, int16))

    if (fileMode .eq. 'create') then
      ! Open file for write:
      newListSet = dataSet(newListSetName, CREATE)

      call writeHeaderKeywords(newListSet, instrumentName, attitude)

      newListTab = addTable(newListSet, 'SRCLIST', int(actualNumSrcs))

      call setAttribute(newListTab, 'PRIMESRC', newPrimeSrcIndex)
      if (withEpicSet) then
        write(epicExpIdAttrName, '(a6, i2)') 'E_EXPR', newEpicAttrNum
        write(epicContentAttrName, '(a6, i2)') 'E_CONT', newEpicAttrNum
        write(epicBandAttrName, '(a6, i2)') 'E_BAND', newEpicAttrNum
        write(epicFilterAttrName, '(a6, i2)') 'E_FILT', newEpicAttrNum
        call setAttribute(newListTab, epicExpIdAttrName, epicExpId)
        call setAttribute(newListTab, epicContentAttrName, epicContent)
        call setAttribute(newListTab, epicBandAttrName, epicBandNum)
        call setAttribute(newListTab, epicFilterAttrName, epicFilter)
      end if

      call writeNewTable(newListTab, newListSrcData, actualNumSrcs)

      call release(newListSet)

    else ! fileMode .eq. 'modify'
!print *, 'about to open data set ', oldListSetName
      ! Open file for modify:
      oldListSet = dataSet(oldListSetName, MODIFY)
!print *, 'opened it successfully'
      if (changeAttitude) then
        call deleteRegionExtensions(oldListSet)
        newListSrcData%process = .false.
        newListSrcData%bkgExclude = .false.

        call setAttribute(newListSet, 'RA_PNT',  attitude%ra   / piOn180)
        call setAttribute(newListSet, 'DEC_PNT', attitude%dec  / piOn180)
        call setAttribute(newListSet, 'PA_PNT',  attitude%apos / piOn180)
      end if

!print *, 'passed if changeAttitude'

      oldListTab = table(oldListSet, 'SRCLIST')

      if (changePrimeSrc) then
        call setAttribute(newListTab, 'PRIMESRC', newPrimeSrcIndex)
      end if
      if (withEpicSet) then
        write(epicExpIdAttrName, '(a6, i2)') 'E_EXPR', newEpicAttrNum
        write(epicContentAttrName, '(a6, i2)') 'E_CONT', newEpicAttrNum
        write(epicBandAttrName, '(a6, i2)') 'E_BAND', newEpicAttrNum
        write(epicFilterAttrName, '(a6, i2)') 'E_FILT', newEpicAttrNum
        call setAttribute(newListTab, epicExpIdAttrName, epicExpId)
        call setAttribute(newListTab, epicContentAttrName, epicContent)
        call setAttribute(newListTab, epicBandAttrName, epicBandNum)
        call setAttribute(newListTab, epicFilterAttrName, epicFilter)
      end if

!print *, 'passed if changePrimeSrc, withEpicSet'
      call deleteRows(oldListTab, 0, numberOfRows(oldListTab) - 1)
!print *, 'deleted rows'
      call insertRows(oldListTab, 0, int(actualNumSrcs) - 1)
!print *, 'inserted rows'
      call writeOldTable(oldListTab, newListSrcData, actualNumSrcs)
!print *, 'table written'

      call release(oldListSet)
    end if

  end subroutine rgssources

  !---------------------------------------------------------------------
  subroutine myinit
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

!    pi = 4.0 * atan2(1.0, 1.0) ! rgslib declares this variable.
    twoPi = 2.0 * pi
    piOn180 = pi / 180.0
    rootPi = sqrt(pi)
    root2 = sqrt(2.0)
    root3 = sqrt(3.0)

    proposalLabel(1:8) = 'PROPOSAL'
    onAxisLabel(1:6) = 'ONAXIS'
    ! They are defined this way so they will match with values read from the file.

  end subroutine myinit

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

character(len=labelsize) :: tempUserSrcLabel, tempPrimeSrcLabel
character(len=labelsize-5) :: tempEpicLabelPrefix

    !-------------------------------------------------------------------
    ! Start:

    fileMode = stringParameter('filemode')
    if (fileMode .eq. 'modify') then
      oldListSetName = stringParameter('rgsset')
      changePrimeSrc = booleanParameter('changeprime')
      changeAttitude = booleanParameter('changeattitude')
    else
      newListSetName = stringParameter('rgsset')
      changePrimeSrc = .true.
      instrument     = stringParameter('instrument')
    end if
    clobberOnLabelClash = booleanParameter('clobberonlabel')
    withEpicSet = booleanParameter('withepicset')
    if (withEpicSet) then
      epicSrcSetName = stringParameter('epicset')
!      epicLabelPrefix = stringParameter('epiclabelprefix')
tempEpicLabelPrefix = stringParameter('epiclabelprefix')
epicLabelPrefix(1:len_trim(tempEpicLabelPrefix)) = trim(tempEpicLabelPrefix)
    end if
    addUserSrc = booleanParameter('addusersource')
    if (addUserSrc) then
!      userSrcLabel = ucase(stringParameter('label'))
tempUserSrcLabel = stringParameter('label')
call ucase(tempUserSrcLabel)
userSrcLabel(1:len_trim(tempUserSrcLabel)) = trim(tempUserSrcLabel)
      userSrcRate = realParameter('rate')
      setUserSrcToPrime = booleanParameter('userasprime')
      userSrcStyle = stringParameter('positionstyle')
      if (userSrcStyle .eq. 'radec') then
        userSrcRaDeg = realParameter('ra')
        userSrcDecDeg = realParameter('dec')
      else ! userSrcStyle .eq. 'wrtatt'
        userSrcDispArcmin = realParameter('deltadisp')
        userSrcXdspArcmin = realParameter('deltaxdsp')
      end if
    else
      setUserSrcToPrime = .false.
    end if
    primeSrcStyle = stringParameter('primestyle')
    if (primeSrcStyle .eq. 'label') then
!      primeSrcLabel = ucase(stringParameter('primelabel'))
!primeSrcLabel = stringParameter('primelabel')
!call ucase(primeSrcLabel)
tempPrimeSrcLabel = stringParameter('primelabel')
call ucase(tempPrimeSrcLabel)
primeSrcLabel(1:len_trim(tempPrimeSrcLabel)) = trim(tempPrimeSrcLabel)
    else if (primeSrcStyle .eq. 'index') then
      primeSrcIndex = intParameter('primeindex')
    else if (primeSrcStyle .eq. 'expression') then
      primeSrcExpression = stringParameter('primeexpression')
    end if
    attitudeStyle = stringParameter('attstyle')
    if (attitudeStyle .eq. 'user') then
      userAttitudeRaDeg = realParameter('attra')
      userAttitudeDecDeg = realParameter('attdec')
      userAttitudeAposDeg = realParameter('attapos')
    else if (attitudeStyle .eq. 'mean' .or. attitudeStyle .eq. 'median') then
      atthkSetName = stringParameter('atthkset')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
!  function ucase(arg) result(string)
subroutine ucase(arg)
  !---------------------------------------------------------------------

    implicit none 

!    character(*), intent(in) :: arg
character(*), intent(inout) :: arg
!    character(*) :: string
    integer :: i, asciiIndex, aAsciiIndex, zAsciiIndex, bigAasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    zAsciiIndex = ichar('z')
    bigAasciiIndex = ichar('A')
!    string = arg
    do i = 1, len(arg)
!      asciiIndex = ichar(string(i:i))
asciiIndex = ichar(arg(i:i))
      if (asciiIndex .ge. aAsciiIndex .and. asciiIndex .le. zAsciiIndex) then
!        string(i:i) = char(bigAasciiIndex + asciiIndex - aAsciiIndex)
arg(i:i) = char(bigAasciiIndex + asciiIndex - aAsciiIndex)
      end if
    end do

!  end function ucase
end subroutine ucase

  !---------------------------------------------------------------------
  subroutine getPointing(attitude, instrumentId)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(out) ::&
      attitude ! typedef from oal
    integer(int32), intent(in) ::&
      instrumentId
    real(double) ::&
      startTime
    type(DataSetT) :: atthkSet
    type(ExposureInfoType) ::&
      exposureInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call CAL_setState(instrument=instrumentId)
    call OAL_setState(instrument=instrumentId&
    , datamode=datamodeId('Spectroscopy'))
    call RGSLIB_refreshCache()
    call OAL_exposureInfo(exposureInfo)
    startTime = ToTimeTag(exposureInfo%actualStartTime)

    attitude%ra=0 ; attitude%dec=0; attitude%apos=0; 

    select case (attitudeStyle) 
    case ('start')
      attitude = instrumentPointing(time=startTime)

    case ('user')
      if (userAttitudeRaDeg .lt. 0.0 .or. userAttitudeRaDeg .gt. 360.0) then
call error('illegalAttRa', '')
      end if
      if (userAttitudeDecDeg .lt. -90.0 .or. userAttitudeDecDeg .gt. 90.0) then
call error('illegalAttDec', '')
      end if
      if (userAttitudeAposDeg .lt. 0.0 .or. userAttitudeAposDeg .gt. 360.0) then
call error('illegalAttApos', '')
      end if

      attitude%ra   = userAttitudeRaDeg * piOn180
      attitude%dec  = userAttitudeDecDeg * piOn180
      attitude%apos = userAttitudeAposDeg * piOn180

    case ('mean') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      attitude%ra = real64Attribute(atthkSet,"AAHFRA") * piOn180
      attitude%dec = real64Attribute(atthkSet,"AAHFDEC") * piOn180
      attitude%apos = real64Attribute(atthkSet,"AAHFPA") * piOn180
      call release(atthkSet)

      ! Correct reference frame for boresight:
      attitude = instrumentPointing(attitude, time=startTime)

    case ('median') ! see task atthkgen
      atthkSet = dataSet(atthkSetName , READ)
      attitude%ra = real64Attribute(atthkSet,"MAHFRA") * piOn180
      attitude%dec = real64Attribute(atthkSet,"MAHFDEC") * piOn180
      attitude%apos = real64Attribute(atthkSet,"MAHFPA") * piOn180
      call release(atthkSet)

      ! Correct reference frame for boresight:
      attitude = instrumentPointing(attitude, time=startTime)

    case default
call error ('bug2', '') 
    end select

  end subroutine getPointing

  !---------------------------------------------------------------------
  subroutine initSrcDataStructure(srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(inout) ::&
      srcData

    !-------------------------------------------------------------------
    ! Start:

    srcData%index = myInt16null
    srcData%label = ''
    srcData%raDeg = myReal32null
    srcData%decDeg = myReal32null
    srcData%rate = 0.0
    srcData%deltaDispArcmin = myReal32null
    srcData%deltaXdspArcmin = myReal32null
    srcData%fovPhiDeg = myReal32null
    srcData%fovRdeg = myReal32null
    srcData%confusion = myReal32null
    srcData%process = .false.
    srcData%bkgExclude = .false.
    srcData%fixedOnSky = .true.
    srcData%epicFileAttrNum = myInt32null

  end subroutine initSrcDataStructure

  !---------------------------------------------------------------------
  subroutine fillOutputSrcList(newListSrcData, maxNumSrcs, attitude&
  , lastNewListSrc, newEpicAttrNum)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      maxNumSrcs
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
    type(source_t), dimension(maxNumSrcs), intent(inout) ::&
      newListSrcData
    integer(int16), intent(out) ::&
      newEpicAttrNum,&
      lastNewListSrc
!    character(len=255), intent(out) ::&
!      newEpicTrueFilename
    type(DataSetT) :: oldListSet, epicSrcSet
    type(TableT) :: oldListTab, epicSrcTab
    type(source_t), dimension(:), allocatable ::&
      oldListSrcData, newEpicSrcData
    type(source_t), dimension(1) ::&
      userSrcData
!    logical ::&
!      oldListProposalExists,&
!      oldListOnaxisExists
    integer(int16) ::&
      src,&
      src2,&
      numOldListSrcs,&
      maxNumNewEpicSrcs,&
      lastEpicSrc,&
      firstSrcToCheck

    !-------------------------------------------------------------------
    ! Start:

    call initSrcDataStructure(newListSrcData, maxNumSrcs)

!    do src = 1, maxNumSrcs
!print *, src, newListSrcData(src)%deltaDispArcmin, newListSrcData(src)%raDeg
!    end do
!print *, ''

    select case(fileMode)
    case('create')
      lastNewListSrc = 0
      call addProposalSrc(newListSrcData, maxNumSrcs, lastNewListSrc)!, attitude)
      call addOnaxisSrc(  newListSrcData, maxNumSrcs, lastNewListSrc, attitude)

      newEpicAttrNum = 1

    case('modify')
      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      numOldListSrcs = numberOfRows(oldListTab)
      allocate(oldListSrcData(numOldListSrcs))
      call readTable(oldListTab, oldListSrcData, numOldListSrcs)
      call release(oldListSet)

!call dumpSrcData(oldListSrcData, numOldListSrcs)

      call checkOldList(oldListSrcData, numOldListSrcs)

      lastNewListSrc = 1
      if (changeAttitude) then
        call addOnaxisSrc(oldListSrcData, maxNumSrcs, lastNewListSrc, attitude)
      end if

      ! Load the old list sources into the new array:
      firstSrcToCheck = 1
      call loadAndTestForLabelClashes(oldListSrcData, numOldListSrcs&
      , firstSrcToCheck, newListSrcData, maxNumSrcs, lastNewListSrc)

      newEpicAttrNum&
      = maxval(newListSrcData(1:lastNewListSrc)%epicFileAttrNum) + 1
!*** not very safe because no guarantee null is < 1. Better to cycle through sources and check for nulls one by one.

    case default
call error('bug3', '')
    end select ! fileMode.

!    do src = 1, maxNumSrcs
!print *, src, newListSrcData(src)%deltaDispArcmin, newListSrcData(src)%raDeg
!    end do
!print *, ''

! At present, we have a newList sources array containing at least the PROPOSAL and ONAXIS at indices 1 and 2, and the remainder of the old list if there is one. So far all the labels of the new list entries are guaranteed to be unique. Next job is to load the new epic sources (if withEpicSet) into a separate array.

    if (withEpicSet) then
      epicSrcSet = dataSet(epicSrcSetName, READ)
      epicSrcTab = table(epicSrcSet, 'SRCLIST', READ)
      maxNumNewEpicSrcs = numberOfRows(epicSrcTab)
      allocate(newEpicSrcData(maxNumNewEpicSrcs))

      call readEpicData(epicSrcTab, newEpicSrcData, maxNumNewEpicSrcs, lastEpicSrc)!, newEpicTrueFilename)
      call release(epicSrcSet)

      newEpicSrcData%index = myInt16null
      newEpicSrcData%deltaDispArcmin = myReal32null
      newEpicSrcData%deltaXdspArcmin = myReal32null
      newEpicSrcData%fovPhiDeg = myReal32null
      newEpicSrcData%fovRdeg = myReal32null
      newEpicSrcData%confusion = myReal32null
      newEpicSrcData%process = .false.
      newEpicSrcData%bkgExclude = .false.
      newEpicSrcData%fixedOnSky = .true.
      newEpicSrcData%epicFileAttrNum = newEpicAttrNum

      do src = 1, lastEpicSrc - 1
        do src2 = src + 1, lastEpicSrc
          if (newEpicSrcData(src)%label .eq. newEpicSrcData(src2)%label) then
!print *, src, src2, '*'//newEpicSrcData(src)%label//'*', '*'//newEpicSrcData(src2)%label//'*'
call error('nonUniqueEpicId', '') ! because this means the epic file was screwed-up.
          end if
        end do
      end do
    end if ! withEpicSet

! Get the user source details into a sort of dummy array (this is just so we can use the same sub loadAndTestForLabelClashes() to load the user source):

    if (addUserSrc) call getUserSrc(userSrcData(1))

! Now we add the new epic and user sources (if any) to the output list:

    if (withEpicSet) then
      firstSrcToCheck = 1
!do src = 1, lastEpicSrc
!  print *, src, newEpicSrcData(src)%label
!end do
      call loadAndTestForLabelClashes(newEpicSrcData, lastEpicSrc&
      , firstSrcToCheck, newListSrcData, maxNumSrcs, lastNewListSrc)
!print *, 'after epic loadAndTestForLabelClashes'
!do src = 1, lastNewListSrc
!print *, src, newListSrcData(src)%label
!end do
    end if

    if (addUserSrc) then
      firstSrcToCheck = 1
      call loadAndTestForLabelClashes(userSrcData, int(1, int16)&
      , firstSrcToCheck, newListSrcData, maxNumSrcs, lastNewListSrc)
    end if

! Now the new or output source list is complete, in that all the sources are present. All label values are unique. We have still to fill in or correct a few columns though.

    ! ra/dec or fov:
    ! Note: all sources (except all previous rgs sources in 'modify' mode, and the ONAXIS source in 'create' mode) should have either ra/dec or xdsp/disp/phi/r set to nulls at this point. After the following loop none of these 6 columns should have any nulls in them.
    do src = 1, lastNewListSrc
!print *, src, newListSrcData(src)%deltaDispArcmin, newListSrcData(src)%raDeg
      if (newListSrcData(src)%fixedOnSky) then ! ra and dec should be left
      ! untouched. Do we recalculate the others?
!        if (fileMode .eq. 'create' .or. changeAttitude&
        if (changeAttitude&
        .or. newListSrcData(src)%deltaDispArcmin .eq. myReal32null&
        .or. newListSrcData(src)%deltaXdspArcmin .eq. myReal32null) then ! yes.
          call calculateFOVcoords(newListSrcData(src), attitude)
        end if

      else ! not 'fixed on sky', the disp/xdisp and fov values should not
      ! be touched. Do we recalculate ra and dec?
!print *, 'source', src, 'not fixed on sky.'
!        if (fileMode .eq. 'create' .or. changeAttitude&
        if (changeAttitude&
        .or. newListSrcData(src)%raDeg .eq. myReal32null&
        .or. newListSrcData(src)%decDeg .eq. myReal32null) then ! yes.
          call invertFOVcoords(newListSrcData(src), attitude)
!print *, 'invertFOVcoords() called.'
        end if
      end if
    end do

!    do src = 1, maxNumSrcs
!print *, src, newListSrcData(src)%deltaDispArcmin, newListSrcData(src)%raDeg
!    end do

    ! index:
    do src = 1, lastNewListSrc
      newListSrcData(src)%index = src
    end do

! Note: can't do confusion until after prime source selected.

  end subroutine fillOutputSrcList

  !---------------------------------------------------------------------
  subroutine checkOldList(oldListSrcData, numOldListSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numOldListSrcs
    type(source_t), dimension(numOldListSrcs), intent(in) ::&
      oldListSrcData
    logical ::&
      oldListProposalExists,&
      oldListOnaxisExists
    integer(int16) :: src, indexOfProposal, indexOfOnaxis


    !-------------------------------------------------------------------
    ! Start:

    ! Do some checks of the PROPOSAL source in the old list: 
    oldListProposalExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. proposalLabel) then
!print *, len(oldListSrcData(src)%label), '*'//oldListSrcData(src)%label//'* yes'
        if (oldListProposalExists) then ! there is more than one PROPOSAL
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyProposals', '')

        else
          oldListProposalExists = .true.
          indexOfProposal = src
        end if
!      else
!print *, len(oldListSrcData(src)%label), '*'//oldListSrcData(src)%label//'* NO'
      end if
    end do
if (.not. oldListProposalExists) call error('noProposal', '')
if (indexOfProposal .ne. 1) call error('displacedProposal', '')

    ! Do some checks of the ONAXIS source in the old list: 
    oldListOnaxisExists = .false.
    do src = 1, numOldListSrcs
      if (oldListSrcData(src)%label .eq. onAxisLabel) then
        if (oldListOnaxisExists) then ! there is more than one ONAXIS
        ! in oldListSrcData. Call error, regardless of the state of
        ! clobberOnLabelClash.
call error('tooManyAttSrc', '')

        else
          oldListOnaxisExists = .true.
          indexOfOnaxis = src
        end if
      end if
    end do
if (.not. oldListOnaxisExists) call error('noAttSrc', '')
if (indexOfOnaxis .ne. 2) call error('displacedAttSrc', '')

!****** other checks on old list??? Check labels are unique, and error if not!!
!****** check matchup between PROCESS/BKG_EXCLUDE and the appropriate region extensions!

  end subroutine checkOldList

  !---------------------------------------------------------------------
  subroutine readEpicData(epicSrcTab, newEpicSrcData, maxNumNewEpicSrcs, lastEpicSrc)!, newEpicTrueFilename)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      epicSrcTab
    integer(int16), intent(in) ::&
      maxNumNewEpicSrcs
    type(source_t), dimension(maxNumNewEpicSrcs), intent(inout) ::&
      newEpicSrcData
    integer(int16), intent(out) ::&
      lastEpicSrc
!    character(len=255), intent(out) ::&
!      newEpicTrueFilename
    integer(int16) ::&
      src
    real(single), dimension(:), pointer ::&
      epicRatePtr
    real(double), dimension(:), pointer ::&
      epicRaPtr, epicDecPtr
    integer(int32), dimension(:), pointer ::&
      epicIndexPtr, epicBandPtr

    !-------------------------------------------------------------------
    ! Start:

    call initSrcDataStructure(newEpicSrcData, maxNumNewEpicSrcs)

    epicRaPtr => real64Data(column(epicSrcTab, 'RA'))
    epicDecPtr => real64Data(column(epicSrcTab, 'DEC'))
    if (hasColumn(epicSrcTab, 'ML_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'ML_ID_SRC'))
    else if (hasColumn(epicSrcTab, 'BOX_ID_SRC')) then
      epicIndexPtr => int32Data(column(epicSrcTab, 'BOX_ID_SRC'))
    else
call error('badEpicIdColumn', '')
    end if
    epicBandPtr => int32Data(column(epicSrcTab, 'ID_BAND'))
    epicRatePtr => real32Data(column(epicSrcTab, 'RATE'))

    lastEpicSrc = 0
    do src = 0, maxNumNewEpicSrcs - 1
      if (epicBandPtr(src) .eq. epicBandNum) then
        lastEpicSrc = lastEpicSrc + 1

        newEpicSrcData(lastEpicSrc)%raDeg = real(epicRaPtr(src), single)
        newEpicSrcData(lastEpicSrc)%decDeg = real(epicDecPtr(src), single)
        newEpicSrcData(lastEpicSrc)%rate = epicRatePtr(src)

        write(newEpicSrcData(lastEpicSrc)%label, '(a4, i5.5)') epicLabelPrefix&
        , int(epicIndexPtr(src), int16)
      end if
    end do

!**** get newEpicTrueFilename
!newEpicTrueFilename = "don't know!"

  end subroutine readEpicData

  !---------------------------------------------------------------------
  subroutine getUserSrc(userSrcData)
  !---------------------------------------------------------------------

    implicit none 

    type(source_t), intent(inout) ::&
      userSrcData

    !-------------------------------------------------------------------
    ! Start:

!    if (userSrcLabel .eq. 'PROPOSAL'&
!    .or. userSrcLabel .eq. 'ONAXIS'&
!    .or. userSrcLabel .eq. '') then
!call error('emptyUserLabel', '')
!    end if

    if (userSrcLabel .eq. '') then
call error('emptyUserLabel', '')
    end if

    ! NOTE!! The following checks are designed mostly to detect when the user has allowed the ra, dec, deltadisp etc parameters to take their default values. The user should not do this. The checks depend to some degree on the values of the relevant parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.
    if (userSrcStyle .eq. 'radec') then
      if (userSrcRaDeg .lt. 0.0 .or. userSrcRaDeg .gt. 360.0) then
call error('illegalUserRa', '')
      end if
      if (userSrcDecDeg .lt. -90.0 .or. userSrcDecDeg .gt. 90.0) then
call error('illegalUserDec', '')
      end if

    else ! userSrcStyle .eq. 'wrtatt'
      if (userSrcXdspArcmin .eq. 0.0 .or. userSrcDispArcmin .eq. 0.0) then
        call warning('zeroWrtAttCoord', '')
      end if
    end if

!!!    call initSrcDataStructure(userSrcData, int(1, int16))
!*** change things (maybe make userSrcData an array) so that this can be called instead of the way the defaults are set at present.

    userSrcData%index = myInt16null
    userSrcData%label = userSrcLabel
    if (userSrcStyle .eq. 'radec') then
      userSrcData%raDeg  = userSrcRaDeg
      userSrcData%decDeg = userSrcDecDeg
      userSrcData%deltaDispArcmin = myReal32null
      userSrcData%deltaXdspArcmin = myReal32null
      userSrcData%fovPhiDeg = myReal32null
      userSrcData%fovRdeg   = myReal32null
      userSrcData%fixedOnSky = .true.

    else ! userSrcStyle .eq. 'wrtatt'
      userSrcData%raDeg  = myReal32null
      userSrcData%decDeg = myReal32null
      userSrcData%deltaDispArcmin = userSrcDispArcmin
      userSrcData%deltaXdspArcmin = userSrcXdspArcmin
      call dispXdspToPolar(&
        userSrcData%deltaDispArcmin,&
        userSrcData%deltaXdspArcmin,&
        userSrcData%fovPhiDeg,&
        userSrcData%fovRdeg&
      )
      userSrcData%fixedOnSky = .false.
    end if
    userSrcData%rate = userSrcRate

    userSrcData%confusion = myReal32null
    userSrcData%process = .false.
    userSrcData%bkgExclude = .false.
    userSrcData%epicFileAttrNum = myInt32null

  end subroutine getUserSrc

  !---------------------------------------------------------------------
  subroutine readTable(tab, srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(out) ::&
      srcData
    integer(int32) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(kind=bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr

    !-------------------------------------------------------------------
    ! Start:

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)

    ! write column values to data array:
    srcData(1:numSrcs)%index           = indexColPtr
    srcData(1:numSrcs)%raDeg           = raColPtr
    srcData(1:numSrcs)%decDeg          = decColPtr
    srcData(1:numSrcs)%rate            = rateColPtr
    srcData(1:numSrcs)%deltaDispArcmin = dispColPtr
    srcData(1:numSrcs)%deltaXdspArcmin = xdspColPtr
    srcData(1:numSrcs)%fovPhiDeg       = phiColPtr
    srcData(1:numSrcs)%fovRdeg         = rColPtr
    srcData(1:numSrcs)%confusion       = confnColPtr
    srcData(1:numSrcs)%process         = procColPtr
    srcData(1:numSrcs)%bkgExclude      = excluColPtr
    srcData(1:numSrcs)%fixedOnSky      = fixedColPtr

    do src = 1, numSrcs
      if (isNull(epicnCol, src - 1)) then
        srcData(src)%epicFileAttrNum = myInt32Null
      else
        srcData(src)%epicFileAttrNum = epicnColPtr(src - 1)
      end if

!      call setStringCell(labelCol, src - 1, srcData(src)%label)
      srcData(src)%label = stringCell(labelCol, src - 1)
    end do

  end subroutine readTable

  !---------------------------------------------------------------------
  subroutine addProposalSrc(srcData, numSrcs, lastNewListSrc)!, attitude)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
!    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
    integer(int16), intent(inout) ::&
      lastNewListSrc
    type(source_t), dimension(numSrcs), intent(inout) ::&
      srcData
    type(ProposalInfoType) :: proposalInfo ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    lastNewListSrc = lastNewListSrc + 1
    call OAL_proposalInfo(proposalInfo)
    srcData(lastNewListSrc)%raDeg  = proposalInfo%targetRa
    srcData(lastNewListSrc)%decDeg = proposalInfo%targetDec
!    srcData(lastNewListSrc)%label = "PROPOSAL" 
    srcData(lastNewListSrc)%label = proposalLabel 
!    call calculateFOVcoords(srcData(lastNewListSrc), attitude)

    ! All others: leave as defaults.

  end subroutine addProposalSrc

  !---------------------------------------------------------------------
  subroutine addOnaxisSrc(srcData, numSrcs, lastNewListSrc, attitude)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
    integer(int16), intent(inout) ::&
      lastNewListSrc
    type(source_t), dimension(numSrcs), intent(inout) ::&
!**** should be:    type(source_t), intent(inout) ::&
!**** then call addOnaxisSrc(srcData(lastNewListSrc), attitude)
      srcData

    !-------------------------------------------------------------------
    ! Start:

    lastNewListSrc = lastNewListSrc + 1
!**** when changes above made, move this line to calling routine.
!**** all the above: make similar changes in addProposalSrc().

    srcData(lastNewListSrc)%deltaDispArcmin = 0.0
    srcData(lastNewListSrc)%deltaXdspArcmin = 0.0
!    srcData(lastNewListSrc)%fovRdeg   = 0.0
!    srcData(lastNewListSrc)%fovPhiDeg = 0.0
    call dispXdspToPolar(&
      srcData(lastNewListSrc)%deltaDispArcmin,&
      srcData(lastNewListSrc)%deltaXdspArcmin,&
      srcData(lastNewListSrc)%fovPhiDeg,&
      srcData(lastNewListSrc)%fovRdeg&
    )
!    srcData(lastNewListSrc)%label = 'ONAXIS' 
    srcData(lastNewListSrc)%label = onAxisLabel 
    srcData(lastNewListSrc)%fixedOnSky = .false.
!    call invertFOVcoords(srcData(lastNewListSrc), attitude)
    srcData(lastNewListSrc)%raDeg  = attitude%ra  / piOn180
    srcData(lastNewListSrc)%decDeg = attitude%dec / piOn180

    ! All others: leave as defaults.

  end subroutine addOnaxisSrc

  !---------------------------------------------------------------------
  subroutine dispXdspToPolar(deltaDispArcmin, deltaXdspArcmin&
  , fovPhiDeg, fovRdeg)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in)  :: deltaDispArcmin, deltaXdspArcmin
    real(single), intent(out) :: fovPhiDeg, fovRdeg
    real (single) :: weightedDD
  
    !-------------------------------------------------------------------
    ! Start:

    weightedDD = -deltaDispArcmin / RGSLIB_getFoverL()

    fovPhiDeg = atan2(deltaXdspArcmin, weightedDD) / piOn180
    fovRdeg = sqrt(weightedDD * weightedDD&
    + deltaXdspArcmin * deltaXdspArcmin) / 60.0

  end subroutine dispXdspToPolar

  !---------------------------------------------------------------------
  subroutine loadAndTestForLabelClashes(trialData, numTrialSrcs, firstTrialSrc&
  , firmData, maxNumSrcs, lastFirmSrc)!, message)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numTrialSrcs, firstTrialSrc, maxNumSrcs
    integer(int16), intent(inout) ::&
      lastFirmSrc
    type(source_t), dimension(numTrialSrcs), intent(in) ::&
      trialData
    type(source_t), dimension(maxNumSrcs), intent(inout) ::&
      firmData
!    character(len=*), intent(in) :: message
    logical :: labelFoundBefore
    integer(int16) :: trialSrc, firmSrc
!character(20) :: dummyStr

    !-------------------------------------------------------------------
    ! Start:

!print *, ' '
!do firmSrc = 1, lastFirmSrc
!  print *, firmSrc, firmData(firmSrc)%label
!end do

    do trialSrc = firstTrialSrc, numTrialSrcs
      labelFoundBefore = .false.
      INNER: do firmSrc = 1, lastFirmSrc
        if (trialData(trialSrc)%label .eq. firmData(firmSrc)%label) then
!print *, trialSrc, firmSrc, trialData(trialSrc)%label//'*'//firmData(firmSrc)%label//'*'
          labelFoundBefore = .true.
          write(errstr, '(a)') 'Duplicate label: '//trialData(trialSrc)%label
          if (clobberOnLabelClash) then
            call warning('duplicateLabel')
      exit INNER

          else
call error('duplicateLabel')
!call warning('duplicateLabel')
          end if
        end if
      end do INNER

      if (.not. labelFoundBefore) then
        lastFirmSrc = lastFirmSrc + 1
if (lastFirmSrc .gt. maxNumSrcs) call error('bug5', '')
        firmData(lastFirmSrc) = trialData(trialSrc)
      end if
    end do

!***** if user source is one that clashes, should overwrite the other source, not delete the user source??? probably not.

!***** this thing should return statuses, and let the calling routine do the error/warnings????

  end subroutine loadAndTestForLabelClashes

  !---------------------------------------------------------------------
!  subroutine calculateFOVcoords(newListSrcData, numSrcs, attitude)
  subroutine calculateFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

!    integer(int16), intent(in) ::&
!      numSrcs
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
!    type(source_t), dimension(numSrcs), intent(inout) ::&
!      newListSrcData
type(source_t), intent(inout) ::&
  srcData
    type(SkyPositionType) :: tempArg
!    integer(int16) :: src
    real(double) :: dummy(2)

    !-------------------------------------------------------------------
    ! Start:

!    do src = 1, numSrcs
!      tempArg%ra  = newListSrcData(src)%raDeg  * piOn180
!      tempArg%dec = newListSrcData(src)%decDeg * piOn180
tempArg%ra  = srcData%raDeg  * piOn180
tempArg%dec = srcData%decDeg * piOn180
      dummy = offAxisAngles(SpacecraftAttitude=attitude, SkyPos=tempArg)
!      newListSrcData(src)%deltaDispArcmin = dummy(1) * 60.0 / piOn180
!      newListSrcData(src)%deltaXdspArcmin = dummy(2) * 60.0 / piOn180
srcData%deltaDispArcmin = dummy(1) * 60.0 / piOn180
srcData%deltaXdspArcmin = dummy(2) * 60.0 / piOn180
!***** check that Peter's previous values were correct!!!!

      call dispXdspToPolar(&
!        newListSrcData(src)%deltaDispArcmin,&
!        newListSrcData(src)%deltaXdspArcmin,&
!        newListSrcData(src)%fovPhiDeg,&
!        newListSrcData(src)%fovRdeg&
srcData%deltaDispArcmin,&
srcData%deltaXdspArcmin,&
srcData%fovPhiDeg,&
srcData%fovRdeg&
      )
!    end do

  end subroutine calculateFOVcoords

  !---------------------------------------------------------------------
  subroutine invertFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
    type(source_t), intent(inout) ::&
      srcData
    type(SkyPositionType) :: srcpos
    real(double) :: offaxis(2)

    !-------------------------------------------------------------------
    ! Start:

       offaxis(1) = srcData%deltaDispArcmin * piOn180 / 3600.0
       offaxis(2) = srcData%deltaXdspArcmin * piOn180 / 3600.0
!       srcpos = rgscelestoffset(attitude, offaxis)
       srcpos = rgscelestoffset(attitude, offaxis(1), offaxis(2))
       srcData%raDeg  = srcpos%ra  / piOn180
       srcData%decDeg = srcpos%dec / piOn180

  end subroutine invertFOVcoords

  !---------------------------------------------------------------------
  function calculatePrimeSourceIndex(srcData, numSrcs) result(indexOfPrime)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(in) ::&
      srcData
    type(DataSetT) :: oldListSet
    type(TableT) :: oldListTab
    integer(int16) ::&
      indexOfPrime, oldIndexOfPrime, proposalIndex, src, dummy(1)
    logical :: notFound

    !-------------------------------------------------------------------
    ! Start:

!print *, fileMode
!print *, changePrimeSrc
!print *, addUserSrc
!print *, setUserSrcToPrime
!print *, primeSrcStyle
!print *, primeSrcLabel
!print *, primeSrcIndex

    ! NOTE!!! The algorithm in this subroutine relies for some extent on the parameter defaults, hence IF THESE ARE CHANGED THIS ALGORITHM MAY ALSO NEED TO BE CHANGED.

    ! Get the previous prime source number if there is one:
    if (fileMode .eq. 'modify') then
      oldListSet = dataSet(oldListSetName, READ)
      oldListTab = table(oldListSet, 'SRCLIST', READ)
      oldindexOfPrime = int16Attribute(oldListTab, 'PRIMESRC')
      call release(oldListSet)
    end if

    ! Get the index of the proposal source (better not assume it is 1):
    notFound = .true.
    do src = 1, numSrcs
      if (srcData(src)%label(1:8) .eq. 'PROPOSAL') then ! this is a messy way to do it, but I can't make the direct comparison between the strings work.
        notFound = .false.
        proposalIndex = srcData(src)%index
    exit
      end if
    end do

if (notFound) call error('bug6', 'No proposal source found')

indexOfPrime = -1 ! an impossible default, to pick up bugs in this routine. Eventually this line (and the tests in the main program) could probably be dispensed with.

    if (fileMode .eq. 'modify' .and. .not. changePrimeSrc) then
      indexOfPrime = oldindexOfPrime

    else ! either fileMode .eq. 'create' .or. changePrimeSrc
      if (addUserSrc .and. setUserSrcToPrime) then
        ! Find user source:
        notFound = .true.
        do src = 1, numSrcs
          if (srcData(src)%label .eq. userSrcLabel) then
            notFound = .false.
            indexOfPrime = srcData(src)%index
        exit
          end if
        end do
if (notFound) call error('bug7', '') ! in current implementation user _label_ should always be present, even if there was a label clash with another src and the _actual_ user source has been discarded. In this case indexOfPrime is _not_ going to point to the user source but the existing source that had the same label value. Hence in no case should this error occur.

      else ! either no user source supplied, or it wasn't requested to be
      ! the prime source. Look now at the dedicated prime source parameters.
        ! Calculate indexOfPrime:
        select case(primeSrcStyle)
        case('label')
          notFound = .true.
          do src = 1, numSrcs
            if (srcData(src)%label .eq. primeSrcLabel) then
              notFound = .false.
              indexOfPrime = src ! same (hopefully) as srcData(src)%index.
          exit
            end if
          end do

          if (notFound) then
            indexOfPrime = proposalIndex
            call warning('primeLabelNotFound', 'Label '//trim(primeSrcLabel)&
            //' not found; prime source set to proposal source.')
          end if

        case('index')
          if (primeSrcIndex .le. numSrcs) then ! here we're assuming that
          ! srcData%index values are consecutive and start at 1;
            indexOfPrime = primeSrcIndex

          else
            indexOfPrime = proposalIndex
            call warning('badPrimeIndex', 'Index **** not found; prime source set to proposal source.')
          end if

        case('expression')
call error('notSupported', 'Sorry, this mode of entering the prime source is not yet supported.')

        case('brightest')
          dummy = maxloc(srcData%rate)
          indexOfPrime = dummy(1)
          if (srcData(indexOfPrime)%rate .le. 0.0) then
            indexOfPrime = proposalIndex
            call warning('noBrightPrime', 'The brightest source has a rate &
            &less than or equal to zero; prime source set to proposal source.')
          end if

        case default
print *, '*'//primeSrcStyle//'*'
call error('badPrimeStyle', 'The value of primestyle is not recognized.')
        end select ! primeSrcStyle
      end if ! addUserSrc .and. setUserSrcToPrime
    end if ! fileMode .eq. 'modify' .and. .not. changePrimeSrc

    if (addUserSrc .and. setUserSrcToPrime .and. .not. changePrimeSrc) then
      call warning('noChangePrime', 'You have requested that the user &
      &source be the prime source, but have not enabled a change in &
      &prime source via the parameter changeprime. Prime source not &
      &changed.')
    end if

    if (fileMode .eq. 'modify' .and. changePrimeSrc) then
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        call warning('primeAtLabelDefault', 'You have requested that the &
        &prime source be changed via the parameter changeprime, but you &
        &may have forgotten to specify the new prime source index, because &
        &it is still set to its default setting of *************proposalIndex.')
      end if

      if (oldindexOfPrime .eq. indexOfPrime) then
        call warning('primeAtIndexDefault', 'You have requested that the &
        &prime source be changed via the parameter changeprime, but you &
        &may have forgotten to specify the new prime source index, because &
        &the new prime source index of *********** is the same as the old one.')
      end if

    else
      if (primeSrcStyle .eq. 'label' .and. primeSrcLabel .eq. 'PROPOSAL') then
        call warning('primeAtDefault', 'Just a reminder that you may want &
        &to specify the new prime source index - it is still set to its &
        &default setting of *************proposalIndex.')
      end if
    end if

  end function calculatePrimeSourceIndex

  !---------------------------------------------------------------------
  subroutine calculateConfusion(srcData, numSrcs, primeSrc)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    integer(int16), intent(in) ::&
      primeSrc
    type(source_t), dimension(numSrcs), intent(inout) ::&
      srcData
    integer(int16) ::&
      src
    real(single) :: delta_disp, delta_xdisp, iPrime
 
!****** these should come from CAL (CCF MiscData)
    real, parameter :: p1 = -0.00653
    real, parameter :: p2 =  1.01
    real, parameter :: a1 = 36.9
    real, parameter :: a2 = 593.4
    real, parameter :: b1 = 2.29
    real, parameter :: b2 = 23.3

    ! See package documentation for equation 
    ! offaxisis angle converted to arcsec

    !-------------------------------------------------------------------
    ! Start:

    if (srcData(primeSrc)%rate .le. 0.0) then
      call warning('badPrimeRate', 'The prime source (index *****) has &
      &a RATE value of <= zero. Confusion values cannot be calculated.')

      srcData%confusion = 0.0

    else
      do src = 1, numSrcs
        delta_disp  = abs(srcData(src)%deltaDispArcmin&
        - srcData(primeSrc)%deltaDispArcmin)
        delta_xdisp = abs(srcData(src)%deltaXdspArcmin&
        - srcData(primeSrc)%deltaXdspArcmin)

        iPrime = p1 + p2 * srcData(src)%rate&
        * exp(-0.5*((delta_disp  * 60.0 + a1) / a2)**2)&
        * exp(-0.5*((delta_xdisp * 60.0 + b1) / b2)**2)
        srcData(src)%confusion = iPrime / srcData(primeSrc)%rate
      end do
    end if

  end subroutine calculateConfusion

  !---------------------------------------------------------------------
  subroutine writeHeaderKeywords(set, instrumentName, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    character(len=5), intent(in) :: instrumentName
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal
    type(ExposureInfoType) :: exposureInfo
    type(ProposalInfoType) :: proposalInfo
    type(OdfInfoType)      :: odfInfo
!    character(len=16)      :: filterName

    !-------------------------------------------------------------------
    ! Start:

    call OAL_odfInfo(odfInfo) 
    call OAL_proposalInfo(proposalinfo)
    call OAL_exposureInfo(exposureInfo) 
!    filterName = OAL_getIPPVstring("FILTER")
!**** this generates a warning - find out why.

    call setAttribute(set, 'TELESCOP', 'XMM')
    call setAttribute(set, 'INSTRUME', instrumentName)
!    call setAttribute(set, 'DATAMODE', *****&
!    , comment="Instrument mode (IMAGING, TIMING, FAST, etc.")
!    call setAttribute(set, "FILTER", filterName, comment="Filter ID")
    if (odfInfo%isSlew) then
      call setAttribute(set, "OBS_MODE", "SLEW"&
      , comment="Observation mode (pointing or slew)")
    else
      call setAttribute(set, "OBS_MODE", "POINTING"&
      , comment="Observation mode (pointing or slew)")
    endif
    call setAttribute(set, 'OBS_ID', trim(odfInfo%observationId)&
    , comment="Observation identifier")
    call setAttribute(set, 'EXP_ID', exposureInfo%number&
    , comment="Exposure identifier")
    call setAttribute(set, 'ORIGIN', 'Leicester/SSC'&
    , comment="Origin of FITS file")
    call setAttribute(set, "OBJECT", trim(proposalInfo%targetName)&
    , comment="Name of observed object")
    call setAttribute(set, 'RA_PNT',  attitude%ra   / piOn180)
    call setAttribute(set, 'DEC_PNT', attitude%dec  / piOn180)
    call setAttribute(set, 'PA_PNT',  attitude%apos / piOn180)
    call setAttribute(set, 'RADECSYS', 'FK5')
    call setAttribute(set, 'EQUINOX', '2000.0')
!    call setAttribute(set, 'EXPOSURE',)
!    call setAttribute(set, 'DATE-OBS',)
!    call setAttribute(set, 'DATE-END',)
    call setAttribute(set, "OBSERVER", trim(proposalInfo%GoName)&
    , comment="Name of observer")

!    call setAttribute(set, 'OCB',) ! special RGS header kwd: on-chip binning

  end subroutine writeHeaderKeywords

  !---------------------------------------------------------------------
  subroutine writeNewTable(tab, srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(in) ::&
      srcData
    integer(int16) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr
integer :: dims(1)
!character(20) :: dummyStr

    !-------------------------------------------------------------------
    ! Start:
dims(1) = int(labelsize)
    ! Set up columns:
    indexCol = addColumn(tab, 'INDEX',        INTEGER16)
    labelCol = addColumn(tab, 'LABEL',        STRING, dimensions=dims&
      , comment = 'This string is unique for each source.')
    raCol    = addColumn(tab, 'RA',           REAL32, 'degrees')
    decCol   = addColumn(tab, 'DEC',          REAL32, 'degrees')
    rateCol  = addColumn(tab, 'RATE',         REAL32, 'cts/s',&
      comment = 'Taken from EPIC source list, or user-supplied.')
    dispCol  = addColumn(tab, 'DELTA_DISP',   REAL32, 'arcmin',&
      comment = 'Offset from pointing: dispersion direction.')
    xdspCol  = addColumn(tab, 'DELTA_XDSP',   REAL32, 'arcmin',&
      comment = 'Offset from pointing: cross-dispersion direction.')
    phiCol   = addColumn(tab, 'FOV_PHI',      REAL32, 'degrees',&
      comment = 'Offset from pointing in polar coordinates: angle from the &
      &-ve dispersion axis towards the +ve cross-dispersion axis.')
    rCol     = addColumn(tab, 'FOV_R',        REAL32, 'degrees',&
      comment = 'Offset from pointing in polar coordinates: radius from pointing.')
    confnCol = addColumn(tab, 'CONFUSION',    REAL32,&
      comment = 'Fraction giving a measure of the confusion between each &
      &source and the prime source.')
    procCol  = addColumn(tab, 'PROCESS',      BOOLEAN,&
      comment = 'True means a spectrum extraction region should be constructed &
      &for this source.')
    excluCol = addColumn(tab, 'BKG_EXCLUDE',  BOOLEAN,&
      comment = 'True means this source should be excluded from the &
      &background spectrum extraction region.')
    fixedCol = addColumn(tab, 'FIXED_ON_SKY', BOOLEAN,&
      comment = 'True for sources whose positions were initially entered &
      &as ra and dec.')
    epicnCol = addColumn(tab, 'EPIC_FILE',    INTEGER32,&
      comment = 'Key to identify the appropriate EPICFILEn and EPICBANDn attributes.')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', '')
    end if

!print *, 'within writeNewTable'
!do src = 1, numSrcs
!print *, src, srcData(src)%label, len(srcData(src)%label)&
!, len(trim(srcData(src)%label))
!end do

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate
      dispColPtr(src - 1)  = srcData(src)%deltaDispArcmin
      xdspColPtr(src - 1)  = srcData(src)%deltaXdspArcmin
      phiColPtr(src - 1)   = srcData(src)%fovPhiDeg
      rColPtr(src - 1)     = srcData(src)%fovRdeg
      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

!dummyStr = srcData(src)%label
!dummyStr='epic00003'
!write(dummyStr, '(a4, i5.5)') epicLabelPrefix, 4
!print *, dummyStr
      call setStringCell(labelCol , src - 1, srcData(src)%label)
!      call setStringCell(labelCol , src - 1, dummyStr)
    end do

  end subroutine writeNewTable

  !---------------------------------------------------------------------
  subroutine writeOldTable(tab, srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(in) ::&
      srcData
    integer(int16) ::&
      src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr
!integer :: dims(1)

    !-------------------------------------------------------------------
    ! Start:

    ! Open the columns:
    indexCol = column(tab, 'INDEX')
    labelCol = column(tab, 'LABEL')
    raCol    = column(tab, 'RA')
    decCol   = column(tab, 'DEC')
    rateCol  = column(tab, 'RATE')
    dispCol  = column(tab, 'DELTA_DISP')
    xdspCol  = column(tab, 'DELTA_XDSP')
    phiCol   = column(tab, 'FOV_PHI')
    rCol     = column(tab, 'FOV_R')
    confnCol = column(tab, 'CONFUSION')
    procCol  = column(tab, 'PROCESS')
    excluCol = column(tab, 'BKG_EXCLUDE')
    fixedCol = column(tab, 'FIXED_ON_SKY')
    epicnCol = column(tab, 'EPIC_FILE')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', '')
    end if

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate
      dispColPtr(src - 1)  = srcData(src)%deltaDispArcmin
      xdspColPtr(src - 1)  = srcData(src)%deltaXdspArcmin
      phiColPtr(src - 1)   = srcData(src)%fovPhiDeg
      rColPtr(src - 1)     = srcData(src)%fovRdeg
      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

      call setStringCell(labelCol , src - 1, srcData(src)%label)
    end do

  end subroutine writeOldTable

  !---------------------------------------------------------------------
  subroutine deleteRegionExtensions(set)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
!    type(BlockT) :: block
    integer :: blockNumber

    !-------------------------------------------------------------------
    ! Start:

    do blockNumber = 1, numberOfBlocks(set) - 1 ! start from 1 not 0 because
    ! we don't want to delete the source list!
      call deleteBlock(set, blockNumber)
    end do

  end subroutine deleteRegionExtensions









  !---------------------------------------------------------------------
  subroutine selectSrcs()
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine selectSrcs

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  !---------------------------------------------------------------------
  subroutine dumpSrcData(srcData, numSrcs)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in) ::&
      numSrcs
    type(source_t), dimension(numSrcs), intent(in) ::&
      srcData
    integer :: src

    !-------------------------------------------------------------------
    ! Start:

    do src = 1, numSrcs
      print *, ''
      print *, ''
      print *, 'Source ', src, ':'
      print *, ''
      print *, '       index:', srcData(src)%index
      print *, '       label:', srcData(src)%label
      print *, '          ra:', srcData(src)%raDeg
      print *, '         dec:', srcData(src)%decDeg
      print *, '        rate:', srcData(src)%rate
      print *, '  delta_disp:', srcData(src)%deltaDispArcmin
      print *, '  delta_xdsp:', srcData(src)%deltaXdspArcmin
      print *, '     fov_phi:', srcData(src)%fovPhiDeg
      print *, '       fov_r:', srcData(src)%fovRdeg
      print *, '   confusion:', srcData(src)%confusion
      print *, '     process:', srcData(src)%process
      print *, ' bkg_exclude:', srcData(src)%bkgExclude
      print *, 'fixed_on_sky:', srcData(src)%fixedOnSky
      print *, '    epic_num:', srcData(src)%epicFileAttrNum
    end do

  end subroutine dumpSrcData

  !---------------------------------------------------------------------
  subroutine generic
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine generic
end module rgssources_mod
